//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The transcript window starts and monitors a running Csound process.

#ifndef CSTRANSCRIPTWINDOW_H
#define CSTRANSCRIPTWINDOW_H

#include <Looper.h>
#include <OS.h>
#include <Point.h>
#include <Window.h>

class CsoundAbstractJob;

class CsoundTranscriptWindow : public BWindow
{
	typedef BWindow inherited;
public:
	
	CsoundTranscriptWindow(BPoint origin, const char *pTitle, BLooper &target,
								  const char *pExecutable,
								  const CsoundAbstractJob &job);
	virtual ~CsoundTranscriptWindow(void);

	// Message codes.
	enum {
		kJobFailed			= 'fail',		// failed to start up
		kJobFinished		= 'jfsh',		// finished job
		kAbortJob			= 'abrt'			// abort job (sent from looper)
	};
		
	// Hook functions.
	virtual void MessageReceived(BMessage *pMsg);
	virtual bool QuitRequested(void);
	
private:
	// Unimplemented copy constructor and assignment operator.
	CsoundTranscriptWindow(const CsoundTranscriptWindow &);
	CsoundTranscriptWindow &operator =(const CsoundTranscriptWindow &);
	
	// Open a pseudo-terminal.
	bool OpenPTY(int &pty, int &tty);
	
	// Monitor the background process.
	static int32 MonitorStub(void *pData);
	int32 Monitor(void);
	
	// Kill the background process.
	void KillChild(void);

	// Static data members.
	static int smBadStdin;	// standard input for child process
	
	// Data members.
	BLooper &mTarget;			// controlling Looper
	thread_id mChild;			// ID of child running Csound
	sem_id mChildFinished;	// synchronization semaphore
};

#endif // CSTRANSCRIPTWINDOW_H
